import 'dart:collection';
import 'dart:convert';

import 'package:contentsquare/contentsquare.dart';
import 'package:contentsquare/src/modules/logger/logger.dart';
import 'package:contentsquare/src/modules/screengraph/compression.dart';
import 'package:contentsquare/src/modules/screengraph/render_object_extension.dart';
import 'package:contentsquare/src/modules/webview_tracking/javascript_event_handler.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'package:webview_flutter/webview_flutter.dart';

import '../logger/logger.dart';
import '../screengraph/hit_test/get_element_data_of_hit_render_object.dart';
import '../screengraph/hit_test/hit_element_data.dart';
import 'exceptions/exceptions.dart';
import 'logs/logs.dart';
import 'screengraph_retrieval.dart';





@visibleForTesting
String get a218 {
  switch (defaultTargetPlatform) {
    case TargetPlatform.android:
      return 'true';
    case TargetPlatform.iOS:
      return '1';
    default:
      throw Exception('Unknown platform');
  }
}

/// Widget used to track a WebView.
///
/// {@template contentsquare.webview.webview_tracker_builder_usage}
/// For the tracking to work, you need to:
///   - Set [WebView.javascriptMode] to [JavascriptMode.unrestricted]
///   - Use [ContentsquareWebViewTracker.onWebViewCreated] in [WebView.onWebViewCreated].
///   - Use [ContentsquareWebViewTracker.onPageFinished] in [WebView.onPageFinished].
///
/// Example:
/// ```dart
/// ContentsquareWebViewTrackerBuilder(
///   builder: (context, tracker) {
///     return WebView(
///       initialUrl: '...',
///       javascriptMode: JavascriptMode.unrestricted,
///       onWebViewCreated: tracker.onWebViewCreated,
///       onPageFinished: tracker.onPageFinished,
///     );
///   },
/// );
/// ```
///
/// **Important**: Widgets created between [ContentsquareWebViewTrackerBuilder]
/// and WebView will **not** be tracked:
///   - They will not be selectable in the zoning analysis.
///   - Gestures on those widgets will not be tracked.
/// Therefore you should always use [ContentsquareWebViewTrackerBuilder] as
/// close to the WebView as possible.
/// {@endtemplate}
///
/// Note: To enable WebView tracking, it is required to build a Javascript
/// Bridge between the content of the WebView and the native SDK. To do so,
/// you will have to implement the Contentsquare WebView Javascript Tracking
/// Tag in the web pages called in your app WebViews.
///   - See https://docs.contentsquare.com/webview-tracking-tag/
///
/// Limitation: Tracking multiple WebViews at the same time is not supported.
class ContentsquareWebViewTrackerBuilder extends StatefulWidget {
  /// Creates a widget which provides a tracker to track a WebView.
  const ContentsquareWebViewTrackerBuilder({
    Key? key,
    required this.builder,
  }) : super(key: key);

  /// Callback in which the WebView should be built.
  ///
  /// {@macro contentsquare.webview.webview_tracker_builder_usage}
  final Widget Function(
    BuildContext context,
    ContentsquareWebViewTracker tracker,
  ) builder;

  @override
  State<ContentsquareWebViewTrackerBuilder> createState() =>
      _a136();
}


class _a136
    extends State<ContentsquareWebViewTrackerBuilder> {
  
  late final a77 = a139(
    a74: a362,
    a73: _a219,
  );

  
  String? _a219() {
    
    
    
    
    if (!mounted) {
      a189.a39(a179());
      return null;
    }

    
    
    
    
    
    
    final a314 = context as Element;

    
    
    
    
    
    
    
    
    
    
    final a315 = a314.renderObject;
    if (a315 == null) {
      a189.a39(a181());
      return null;
    }
    if (a315 is! a138) {
      a189.a35(
        'The type of render object associated with the $ContentsquareWebViewTrackerBuilder widget is ${a315.runtimeType}.\n'
        'This is an error, it should always be a $a138.\n',
      );
      return null;
    }

    
    
    
    
    
    
    
    
    final a316 = a315.a232;
    if (a316 == null) {
      a189.a39(a180());
      return null;
    }

    final a317 = a269(
      a530: a315,
      a531: a142.a238(
        a500: a316.center,
      ),
    );

    return a317.a14(
      a426: (hitElementData) {
        return a195(hitElementData.a67);
      },
      
      
      a425: () => null,
    );
  }

  @override
  void dispose() {
    a77._a229();
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return _a140(
      a493: _a137(
        a485: widget.builder(context, a77),
        a76: a77.a23,
      ),
    );
  }
}


class _a137
    extends SingleChildRenderObjectWidget {
  const _a137({
    Key? a484,
    required this.a76,
    required Widget a485,
  }) : super(key: a484, child: a485);

  
  final Future<Map<String, dynamic>> Function() a76;

  @override
  RenderObject createRenderObject(BuildContext context) {
    return a138(a75: a76);
  }
}


class a138 extends RenderProxyBox {
  
  a138({
    required this.a75,
  });

  
  final Future<Map<String, dynamic>> Function() a75;
}

/// An object used to track a WebView.
///
/// {@macro contentsquare.webview.webview_tracker_builder_usage}
abstract class ContentsquareWebViewTracker {
  /// Method to use in the [WebView.onWebViewCreated] callback.
  ///
  /// Example:
  /// ```dart
  /// WebView(
  ///   ...
  ///   onWebViewCreated: tracker.onWebViewCreated,
  /// )
  /// ```
  ///
  /// If you need to use the [WebView.onWebViewCreated] callback, you can do
  /// the following:
  /// ```dart
  /// WebView(
  ///   ...
  ///   onWebViewCreated: (controller) {
  ///     tracker.onWebViewCreated(controller);
  ///
  ///     // Do whatever you need to do.
  ///   },
  /// )
  /// ```
  void onWebViewCreated(WebViewController controller);

  /// Method to use in the [WebView.onPageFinished] callback.
  ///
  /// Example:
  /// ```dart
  /// WebView(
  ///   ...
  ///   onPageFinished: tracker.onPageFinished,
  /// )
  /// ```
  ///
  /// If you need to use the [WebView.onPageFinished] callback, you can do
  /// the following:
  /// ```dart
  /// WebView(
  ///   ...
  ///   onPageFinished: (url) async {
  ///     await tracker.onPageFinished(url);
  ///
  ///     // Do whatever you need to do.
  ///   },
  /// )
  /// ```
  Future<void> onPageFinished(String url);

  /// Set of javascript channels to register in the [WebView.javascriptChannels] callback.
  ///
  /// Example:
  /// ```dart
  /// WebView(
  ///   ...
  ///   javascriptChannels: tracker.javascriptChannels,
  /// )
  /// ```
  ///
  /// If you need to add custom javascript channels in the [WebView], you can do
  /// the following:
  /// ```dart
  /// WebView(
  ///   ...
  ///   javascriptChannels: {
  ///      ...tracker.javascriptChannels,
  ///
  ///     // Add you javascript channels.
  ///   },
  /// )
  /// ```
  Set<JavascriptChannel> get javascriptChannels;
}


class a139 implements ContentsquareWebViewTracker {
  
  a139({
    required this.a74,
    required this.a73,
  });

  
  final a146 a74;

  
  final String? Function() a73;

  
  WebViewController? _a318;

  @override
  void onWebViewCreated(WebViewController controller) {
    _a318 = controller;
  }

  @override
  Future<void> onPageFinished(String url) async {
    final _a319 = this._a318;

    if (_a319 == null) {
      return a189.a39(
        a158(
          a527: null,
          a528:
              'WebViewController is null, did you forget to use tracker.onWebViewCreated?',
        ),
      );
    }

    final bool a320;
    try {
      a320 = await _a221(
        a486: _a319,
      );
    } catch (error) {
      return a189.a39(
        a158(
          a527: await _a319.currentUrl(),
          a528:
              'Unexpected error while executing JS code, check the registration of the javascript channels: $error',
        ),
      );
    }
    if (!a320) {
      return a189.a39(
        a158(
          a527: await _a319.currentUrl(),
          a528:
              'JavascriptChannels are not registered, did you forget to use tracker.javascriptChannels?',
        ),
      );
    }

    try {
      await _a319.runJavascript(_a220);
    } catch (error) {
      return a189.a39(
        a158(
          a527: await _a319.currentUrl(),
          a528:
              'Unexpected error while executing JS code to setup the Javascript bridge: $error',
        ),
      );
    }

    return a189.a39(
      a160(
        a529: await _a319.currentUrl(),
      ),
    );
  }

  @override
  Set<JavascriptChannel> get javascriptChannels {
    return ContentsquareJavascriptEventEnum.values
        .map(
          (event) => a149(
            a57: Contentsquare(),
            a56: event,
            a55: a73,
          ).a2,
        )
        .toSet();
  }

  
  String get _a220 {
    return '''
window.CSJavascriptBridge = {
  
  ${ContentsquareJavascriptEventEnum.values.map(
              (event) => '${event.name}: ${a149(
                a57: Contentsquare(),
                a56: event,
                a55: a73,
              ).a1}',
            ).join(',\n\n').replaceAll('\n', '\n  ')}
};''';
  }

  
  Future<bool> _a221({
    required WebViewController a486,
  }) async {
    for (final a321 in javascriptChannels) {
      final a322 =
          await _a222(
        a321,
        a488: a486,
      );

      if (!a322) {
        return false;
      }
    }

    return true;
  }

  Future<bool> _a222(
    JavascriptChannel a487, {
    required WebViewController a488,
  }) async {
    return await a488.runJavascriptReturningResult(
          _a223(a487),
        ) ==
        a218;
  }

  String _a223(JavascriptChannel a489) =>
      '''window.${a489.name} !== undefined''';

  
  
  
  
  var _a323 = UnmodifiableListView<Stopwatch>([]);

  
  Stopwatch _a224() {
    final a324 = Stopwatch();
    _a323 = UnmodifiableListView(_a323 + [a324]);
    return a324;
  }

  
  void _a225(Stopwatch a490) {
    a490.stop();
    _a323 = UnmodifiableListView(
      _a323.toList()..remove(a490),
    );
  }

  
  static String _a226(String a491) => '''
    if (window.cs_wvt == undefined || window.cs_wvt == null) {
      // Cast on 'null' for jsonDecode to decode it to dart null.
      'null';
    } else {
      $a491;
    }
  ''';

  
  @visibleForTesting
  static final a72 =
      _a226("JSON.parse(cs_wvt.push(['serializeWebView']));");

  
  @visibleForTesting
  static final a71 =
      _a226("window.cs_wvt.push(['serializeWebView']);");

  
  @visibleForTesting
  static String get a6 {
    switch (defaultTargetPlatform) {
      case TargetPlatform.android:
        return a72;
      case TargetPlatform.iOS:
        return a71;
      default:
        throw a169();
    }
  }

  
  Future<Map<String, dynamic>> a23() async {
    final a325 = _a318;
    if (a325 == null) {
      throw a178();
    }

    
    
    final a326 = _a224();
    a326.start();

    bool a227() {
      return a326.elapsed >=
          a74.a60;
    }

    
    
    
    
    bool a228() {
      return !a326.isRunning;
    }

    Map<String, dynamic>? a327;
    
    
    
    
    while (!a227() && !a228()) {
      final a328 = await a325
          .runJavascriptReturningResult(a6);
      final Object? a329;
      try {
        a329 = jsonDecode(a328);
      } on FormatException catch (e) {
        a189.a39(
          a174(
            a42: a328,
            a41: e,
          ),
        );
        continue;
      }

      
      
      
      
      
      
      
      final a330 = a329 == null ||
          a329 is num ||
          a329 is Map<String, dynamic>;
      if (!a330) {
        a189.a39(
          a173(
            a43: a329,
          ),
        );
      }

      if (a329 is Map<String, dynamic>) {
        a327 = a329;
        break;
      }

      await Future.delayed(
          a74.a59);
    }

    _a225(a326);

    if (a327 == null) {
      throw a175(
        a40: a74.a60,
      );
    }

    return a327;
  }

  
  void _a229() {
    _a323.forEach(_a225);
  }
}


class _a140 extends InheritedWidget {
  const _a140({
    Key? a492,
    required Widget a493,
  }) : super(key: a492, child: a493);

  @override
  bool updateShouldNotify(oldWidget) => false;
}


extension a184 on Element {
  
  bool get _a230 {
    return widget is ContentsquareWebViewTrackerBuilder;
  }

  
  bool get a231 {
    final a331 =
        getElementForInheritedWidgetOfExactType<_a140>();
    return a331 != null || _a230;
  }
}
